from typing import Literal


SOURCE = Literal[
    "ambe",
    "ar-37",
    "kr-83m",
    "rn-220",
    "led",
    "pulser",
    "noise",
    "none",
    "th-228",
    "th-232",
    "Y88",
    "YBe",
]

DIFFUSED_SOURCE = Literal["rn-220", "kr-83m", "ar-37"]

DETECTOR = Literal["tpc", "neutron_veto", "muon_veto", "tpc_he"]

PARTITION = Literal["all_tpc", "ab", "cd"]



"""

"""
import datetime
from typing import Literal
import rframe

from .base_references import BaseResourceReference

class NaiveBayesClassifier(BaseResourceReference):
    _ALIAS = "bayes_models"
    fmt = "binary"
    
    time: rframe.Interval[datetime.datetime] = rframe.IntervalIndex()
    value: str



import pandas as pd
from rframe import BaseSchema


def xent_collection(**kwargs):
    pass

try:
    import utilix

    uconfig = utilix.uconfig
    from utilix import xent_collection

except ImportError:
    uconfig = None

from pydantic import BaseSettings

from .clock import SimpleClock

    
class Settings(BaseSettings):
    class Config:
        env_prefix = "XEDOCS_"

    ANALYST_DB: str = "xedocs-dev"
    STRAXEN_DB: str = "xedocs"
    API_URL_FORMAT: str = "{base_url}/{version}/{mode}/{name}"
    API_BASE_URL: str = "https://api.xedocs.yossisprojects.com"
    API_VERSION: str = "v1"
    API_AUDIENCE: str = "https://api.cmt.xenonnt.org"
    API_READONLY: bool = False
    API_TOKEN: str = None
    API_USERNAME: str = None
    API_PASSWORD: str = None

    clock = SimpleClock()

    datasources = {}

    @property
    def token(self):
        if self.API_TOKEN is None:
            self.login()

        if hasattr(self.API_TOKEN, "access_token"):
            if self.API_TOKEN.expired:
                self.API_TOKEN.refresh()
            return self.API_TOKEN.access_token
        return self.API_TOKEN

    def login(self):
        import xeauth

        if self.API_READONLY:
            scopes = ["read:all"]
        else:
            scopes = ["read:all", "write:all"]
            
        token = xeauth.login(
            username=self.API_USERNAME,
            password=self.API_PASSWORD,
            scopes=scopes,
            audience=self.API_AUDIENCE,
        )

        self.API_TOKEN = token

    def api_url_for_schema(self, schema, base_url=None, version=None, mode='staging'):
        if base_url is None:
            base_url = self.API_BASE_URL
        
        if version is None:
            version = self.API_VERSION

        if hasattr(schema, "_ALIAS"):
            schema = schema._ALIAS

        return self.API_URL_FORMAT.format(
            base_url=base_url.strip('/'), version=version.strip('/'),
            name=schema.strip('/'), mode=mode.strip('/')
        )

    def run_doc(self, run_id, fields=("start", "end")):
        if uconfig is None:
            raise KeyError(f"Rundb not configured.")

        rundb = xent_collection()

        if isinstance(run_id, str):
            run_id = int(run_id)

        query = {"number": run_id}

        doc = rundb.find_one(query, projection={f: 1 for f in fields})
        if not doc:
            raise KeyError(f"Run {run_id} not found.")

        return doc

    def run_id_to_time(self, run_id):
        doc = self.run_doc(run_id)
        # use center time of run
        return doc["start"] + (doc["end"] - doc["start"]) / 2

    def run_id_to_interval(self, run_id):
        doc = self.run_doc(run_id)
        return doc["start"], doc["end"]

    def extract_time(self, kwargs):
        if "time" in kwargs:
            time = kwargs.pop("time")
        elif "run_id" in kwargs:
            time = self.run_id_to_time(kwargs.pop("run_id"))
        else:
            return None
        return pd.to_datetime(time)


settings = Settings()



"""Main module."""

import pandas as pd

from collections import defaultdict
from typing import ClassVar, Dict, List, Type, Union
from rframe import DataAccessor

from ._settings import settings
from .schemas import XeDoc


def find_docs(schema, datasource=None, **labels):
    """find documents by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        List[XeDoc]: a list of documents matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_docs(**labels)


def find_iter(schema, datasource=None, **labels):
    """find documents by labels, return iterator

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        Iterator[XeDoc]: an iterator over documents matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_iter(**labels)


def find_df(schema, datasource=None, **labels):
    """find dataframe by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        DataFrame: a pandas dataframe matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_df(**labels)


def find_one(schema, datasource=None, **labels):
    """find dataframe by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        Optional[XeDoc]: The first document matching selection or None.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_one(**labels)


def insert_docs(schema: str, docs: Union[list, dict, pd.DataFrame], datasource=None):
    if isinstance(docs, pd.DataFrame):
        docs = docs.reset_index().to_dict(orient="records")
    if not isinstance(docs, list):
        docs = [docs]
    accessor = get_accessor(schema, datasource)

    return accessor.insert(docs)


def list_schemas() -> List[str]:
    return list(XeDoc._XEDOCS)


def all_schemas() -> Dict[str, Type[XeDoc]]:
    return dict(XeDoc._XEDOCS)


def schemas_by_category() -> Dict[str,Dict[str, Type[XeDoc]]]:
    d = defaultdict(dict)
    for name, schema in all_schemas().items():
        d[schema._CATEGORY][name] = schema
    return d


def find_schema(name) -> Type[XeDoc]:
    if issubclass(name, XeDoc):
        return name

    schema = XeDoc._XEDOCS.get(name, None)

    if schema is not None:
        return schema
    
    for schema in XeDoc._XEDOCS.values():
        if schema.__name__ == name:
            return schema
    
    raise KeyError(f"Correction with name {name} not found.")


def get_accessor(schema, db='analyst_db'):
    schema = find_schema(schema)

    if not issubclass(schema, XeDoc):
        raise TypeError(
            "Schema must be a subclass of XeDoc" "or the name of a known schema."
        )
    if db is None:
        db = 'analyst_db'
    if isinstance(db, str):
        accessor = getattr(schema, db)
    else:
        accessor = DataAccessor(schema, db)

    return accessor


def help(schema):
    if isinstance(schema, str):
        schema = find_schema(schema)
    return schema.help()


def get_api_client(schema):
    if isinstance(schema, str):
        schema = find_schema(schema)
    return settings.api_client(schema)




import pandas as pd

from collections import UserDict


def docs_to_wiki(schema, docs, title=None, columns=None):
    """Convert a list of documents to a dokuwiki table

    :param title: title of the table.
    """
    if title is None:
        title = schema._ALIAS.replace("_", " ").capitalize() + " Table"

    if columns is None:
        columns = list(schema.get_index_fields()) + list(schema.get_column_fields())

    table = "^ %s " % title + "^" * (len(columns) - 1) + "^\n"
    table += "^ " + " ^ ".join(columns) + " ^\n"

    for doc in docs:
        table += "| " + " | ".join([str(getattr(doc, col)) for col in columns]) + " |\n"
    return table


def docs_to_dataframe(schema, docs, columns=None):
    if columns is None:
        columns = list(schema.__fields__)
    else:
        columns = list(columns)

    docs = [doc.pandas_dict() for doc in docs]
    df = pd.json_normalize(docs)

    if len(df):
        df = df[columns]
    else:
        df = df.reindex(columns=list(columns))

    index_fields = [name for name in schema.get_index_fields() if name in columns]

    if len(index_fields) == 1:
        index_fields = index_fields[0]

    return df.set_index(index_fields)


class DatasetCollection(UserDict):
    def __getattr__(self, attr):
        if attr in self.keys():
            return self[attr]
        raise AttributeError(attr)

    def __dir__(self):
        return list(self.keys())



'''
Correction: Average single electron gain
Affects: Corrected areas

Average of the single electron corrections. This correction my also have two partitions as this average is a temporal average not a spatial one.

wiki:
'''

import rframe

from .base_corrections import TimeSampledCorrection
from ..constants import PARTITION

class AvgSingleElectronGain(TimeSampledCorrection):
    
    _ALIAS = "avg_se_gains"
    partition: PARTITION = rframe.Index(default='all_tpc')
    value: float


"""
# TPC PMT hitfinder thresholds
`hit_thresholds` are the per-channel thresholds in unit of ADC to define a hit in `strax`, and they should be updated while DAQ thresholds and noise condition change. This folder contains essential info and source codes of `straxen` hit thresholds for TPC PMTs (both normal channels and high energy channels) and Neutron Veto PMTs. `hitfinder_thresholds.py` in this folder contains the thresholds of different versions. 

"""

import rframe

from .base_corrections import TimeSampledCorrection
from ..constants import DETECTOR


class HitThreshold(TimeSampledCorrection):
    _ALIAS = "hit_thresholds"
    detector: DETECTOR = rframe.Index()
    pmt: int = rframe.Index()

    value: int



from typing import Union
from abc import ABC, abstractmethod

from .api import ApiAuth
from .schemas import XeDoc
from xedocs import settings


class BaseStorage(ABC):
    @abstractmethod
    def get_datasource(self, name: Union[str, XeDoc]):
        pass


class DictStorage(BaseStorage):
    def __init__(self, datasources=None):
        if datasources is None:
            datasources = {}
        self.datasources = datasources

    def get_datasource(self, name: str):
        return self.datasources.get(name, None)


class MongoDBStorage(BaseStorage):
    def __init__(self, db):
        self.db = db

    def get_datasource(self, name: str):
        return self.db[name]


class UtilixStorage(BaseStorage):
    DB_NAME: str = "xedocs-dev"

    def __init__(self, database=settings.ANALYST_DB):
        self.DB_NAME = database

    def get_datasource(self, name):
        try:
            import utilix
        except ImportError:
            return None
        return utilix.xent_collection(collection=name, database=self.DB_NAME)


class ApiStorage(BaseStorage):
    MODE: str = "anaylst"

    def __init__(self, mode="anaylst"):
        self.MODE = mode

    def get_datasource(self, name: str):
        import xedocs
        return xedocs.api_client(name, mode=self.MODE)




"""Attempt to register URConfig protocol if straxen
is installed.
"""
import straxen


@straxen.URLConfig.register("xedocs")
def xedocs_protocol(
    name, db="straxen_db", sort=None, attr=None, as_list=False, **labels
):
    """URLConfig protocol for fetching values from
        a xedocs database.
    ::param name: Name of the schema.
    ::param context: Context of the document.
    ::param version: Version of the documents to filter by.
    ::param sort: Attribute of the documents to sort on.
    ::param attr: Attribute of the documents to return.
    ::param labels: Label values to filter by to return.
    """
    import xedocs

    # Find the document schema
    schema = xedocs.find_schema(name)

    accessor = getattr(schema, db)

    if sort is not None:
        labels['sort'] = sort

    docs = accessor.find_docs(**labels)

    if not docs:
        raise KeyError(f"No matching documents found for {name}.")

    if isinstance(sort, str):
        docs = sorted(docs, key=lambda x: getattr(x, sort))
    elif sort:
        docs = sorted(docs)

    if attr is not None:
        docs = [getattr(d, attr) for d in docs]

    if not as_list:
        return docs[0]

    return docs



import datetime
import re
from typing import ClassVar, List
from pydantic import validator, BaseModel, Field

import pandas as pd
import rframe

from ..._settings import settings

from ..base_schemas import VersionedXeDoc


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


class Review(BaseModel):
    reviewer: str = Field(max_length=80)
    approved: bool = False
    comments: str = ""


class BaseCorrectionSchema(VersionedXeDoc):
    """Base class for all correction schemas.
    This class ensures:

    - the _ALIAS attribute is always unique
    - schema includes a version index
    - changing already set values is disallowed

    """

    _ALIAS: ClassVar = ""
    _CATEGORY = "corrections"
    _CORRECTIONS = {}

    created_date: datetime.datetime = settings.clock.current_datetime()
    comments: str = ""
    reviews: List[Review] = []

    def __init_subclass__(cls) -> None:

        if cls._ALIAS in cls.__dict__ and cls._ALIAS not in cls._CORRECTIONS:
            cls._CORRECTIONS[cls._ALIAS] = cls

        super().__init_subclass__()

    def pre_update(self, datasource, new):
        """This method is called if the `new` document is
        being saved and self was found to already exist in
        the datasource. By default we check that all values
        are the same. The reason this execption is needed is
        because the found document may not actually exist in
        the datasource and may be interpolated, so we allow
        updating documents with identical values.
        Otherwise we raise an error, preventing the update.
        """

        if not self.same_values(new):
            index = ", ".join([f"{k}={v}" for k, v in self.index_labels.items()])
            raise IndexError(f"Values already set for {index}")

    def pre_delete(self, datasource, **kwargs):
        raise RuntimeError("Corrections are append only.")


class TimeIntervalCorrection(BaseCorrectionSchema):
    """Base class for time-interval corrections

    - Adds an Interval index of type datetime
    - Enforces rules on updating intervals:
    Can only change the right side of an interval
    if right side is None and the new right side is
    after the cutoff time (default is 2 hours after current time).

    The cutoff is set to prevent values changing after already being used
    for processing data.
    """

    class Config:
        allow_population_by_field_name = True

    _ALIAS = ""

    time: rframe.Interval[datetime.datetime] = rframe.IntervalIndex(alias="run_id")

    @validator("time", pre=True)
    def run_id_to_time(cls, v):
        """Convert run id to time"""
        if isinstance(v, (str, int)):
            try:
                v = settings.run_id_to_time(v)
            except:
                pass
        return v

    @classmethod
    def url_protocol(cls, attr, **labels):
        labels["time"] = settings.extract_time(labels)
        return super().url_protocol(attr, **labels)

    def pre_update(self, datasource, new):
        """Since intervals can extend beyond the current time,
        we want to allow changes to the end time shortening the interval
        to a point in the future since these values have not yet
        been used for processing.
        """
        current_left = self.time.left
        current_right = self.time.right
        new_left = new.time.left
        new_right = new.time.right

        clock = settings.clock
        cutoff = clock.cutoff_datetime(buffer=60)

        if clock.after_cutoff(current_left) and clock.after_cutoff(new_left):
            # current and new interval are completely in the future
            # all changes are allowed
            return

        if current_right > new_right:
            # Interval is being shortened.
            # We only allow shortening intervals that extend beyong the cutoff time
            assert clock.after_cutoff(
                current_right
            ), f"Can only shorten intervals \
                                                        that ends after {cutoff}"

            # The resulting interval must extend beyong the cutoff time
            assert clock.after_cutoff(
                new_right
            ), f"Can only shorten an interval \
                                                    to end after {cutoff}"

        # Only allow changes to the right side of the interval
        assert (
            current_left == new_left
        ), f"Can only change endtime of existing interval. \
                                           start time must be {self.time.left}"

        # Only allow changes to the interval, not the values
        assert self.same_values(new), f"Values already set for {self.index_labels}."

    def pre_delete(self, datasource, **kwargs):
        if settings.clock.after_cutoff(self.time.left):
            # We allow deletion of future values for all versions
            # if they are completely in the future
            return
        # all other cases, deletion is forbiden.
        raise RuntimeError("Corrections are append only.")

    @classmethod
    def validity_intervals(cls, datasource=None, **labels):
        """Returns a list of intervals that are valid for the given labels"""
        ivs = cls.unique(datasource, fields="time", **labels)
        ivs = sorted(ivs)
        if not ivs:
            return []
        merged = ivs[:1]
        for iv in ivs[1:]:

            if iv.left == merged[-1].right:
                merged[-1] = merged[-1].clone(right=iv.right)
            else:
                merged.append(iv)
        return merged


def can_extrapolate(doc):
    # only extrapolate ONLINE versions
    # and up until the current time.
    if doc["version"] == "ONLINE":
        ts = pd.to_datetime(doc["time"]).to_pydatetime()
        clock = settings.clock
        return not clock.after_cutoff(ts)

    return False


class TimeSampledCorrection(BaseCorrectionSchema):
    """Base class for time-sampled corrections

    - Adds an interpolating index of type datetime
    - Enforces rules on inserting new data points

    Since extrapolation is allowed for ONLINE versions
    Inserting new points before the cutoff is disallowed
    This is to prevent setting values for times already
    processed using the extrapolated values.
    When inserting an ONLINE value after the cutoff, a
    new document with equal values to the extrapolated values
    is inserted at the current time to prevent the inserted document
    from affecting the interpolated values that have already been used
    for processing.
    """

    class Config:
        allow_population_by_field_name = True

    _ALIAS = ""

    time: datetime.datetime = rframe.InterpolatingIndex(
        extrapolate=can_extrapolate, alias="run_id"
    )

    @validator("time", pre=True)
    def run_id_to_time(cls, v):
        """Convert run id to time"""
        if isinstance(v, (str, int)):
            try:
                v = settings.run_id_to_time(v)
            except:
                pass
        return v

    @classmethod
    def url_protocol(cls, attr, **labels):
        labels["time"] = settings.extract_time(labels)
        return super().url_protocol(attr, **labels)

    def freeze_values(self, datasource):
        new_index = self.index_labels
        new_index["time"] = settings.clock.cutoff_datetime(buffer=1)

        existing = self.find(datasource, **new_index)

        # If values for the cutoff time are already set, the values may have been
        # used for processing. We add a sample at the
        # cutoff time to force interpolation and extrapolation
        # to match from the last existing sample until the cutoff.
        if existing:
            new_doc = existing[0]
            new_doc.save(datasource)

    def pre_insert(self, datasource):
        # Inserting ONLINE versions can affect the past
        # since extrapolation until the current time is allowed
        # extrapolation will just give the last value,
        # verses interpolation which is calculated from the last
        # value and also the newly inserted value.
        # For this reason, when inserting ONLINE versions
        # an additional document is inserted with the current time
        # and values equal to the latest existing document.
        # This sets all values from the latest document time until now
        # permanently to the values that would have been used in processing.
        if self.version == "ONLINE":
            clock = settings.clock
            cutoff = clock.cutoff_datetime(buffer=60)

            assert clock.after_cutoff(
                self.time
            ), f"Can only insert online \
                values after {cutoff}."

            self.freeze_values(datasource)

    def pre_delete(self, datasource, **kwargs):
        cutoff = settings.clock.cutoff_datetime(buffer=60)

        assert settings.clock.after_cutoff(
            self.time
        ), f"Can only delete \
                values after {cutoff}."

        if self.version == "ONLINE":
            # deleting ONLINE values can affect interpolation of
            # older values so we need to freeze the values up
            # until the current time.
            self.freeze_values(datasource)

    @classmethod
    def validity_intervals(cls, datasource=None, **labels):
        left = cls.min(datasource, fields="time", **labels)
        right = cls.max(datasource, fields="time", **labels)
        if left is None or right is None:
            return []
        if "version" not in labels or labels["version"] == "ONLINE":
            right = max(settings.clock.cutoff_datetime(), right)
        return [rframe.Interval[datetime.datetime](left=left, right=right)]



#!/usr/bin/env python
"""Tests for `xedocs` package."""
# pylint: disable=redefined-outer-name

import pytest
from click.testing import CliRunner

from xedocs import cli


def test_analyst_context():
    from xedocs.contexts import analyst_db

    ctx = analyst_db()
    assert ctx is not None
    ctx = analyst_db(by_category=False)
    assert ctx is not None


def test_straxen_context():
    from xedocs.contexts import straxen_db

    ctx = straxen_db()
    assert ctx is not None
    ctx = straxen_db(by_category=False)
    assert ctx is not None


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)



import rframe
import pandas as pd


from typing import Any
from plum import dispatch


@dispatch
def json_serializable(value: Any):
    return value


@dispatch
def json_serializable(value: list):
    return [json_serializable(v) for v in value]


@dispatch
def json_serializable(value: tuple):
    return tuple(json_serializable(v) for v in value)


@dispatch
def json_serializable(value: dict):
    return {json_serializable(k): json_serializable(v) for k, v in value.items()}


@dispatch
def json_serializable(value: pd.Interval):
    return json_serializable((value.left, value.right))


@dispatch
def json_serializable(value: rframe.IntegerInterval):
    return value.left, value.right


@dispatch
def json_serializable(value: rframe.TimeInterval):
    return f"{str(value.left)} to {str(value.right)}"



import datetime
import pydantic
import rframe

try:
    import utilix

    uconfig = utilix.uconfig
    from utilix import xent_collection

except ImportError:
    uconfig = None

    def xent_collection(**kwargs):
        raise RuntimeError("utilix not configured")


from pydantic import validator
from typing import Literal

from .base_analysis import BaseAnalysisSchema
from ..._settings import settings


RSE_TYPE = Literal[
    "SURFSARA_USERDISK",
    "SDSC_USERDISK",
    "LNGS_USERDISK",
    "UC_OSG_USERDISK",
    "UC_DALI_USERDISK",
    "CNAF_USERDISK",
]


def xeauth_user():
    if uconfig is None:
        return "unknown"
    return uconfig.get("xeauth", "api_user", fallback="unknown")


class ProcessingRequest(BaseAnalysisSchema):
    """Schema definition for a processing request"""

    _ALIAS = "processing_requests"

    data_type: str = rframe.Index()
    lineage_hash: str = rframe.Index()
    run_id: str = rframe.Index()
    destination: RSE_TYPE = rframe.Index(default="UC_DALI_USERDISK")
    user: str = pydantic.Field(default_factory=xeauth_user)
    request_date: datetime.datetime = pydantic.Field(
        default_factory=datetime.datetime.utcnow
    )

    priority: int = -1

    comments: str = ""

    def pre_update(self, datasource, new):
        if new.user != self.user:
            raise ValueError(new.user)
        if new.run_id != self.run_id:
            raise ValueError(new.run)

    def latest_context(self):
        import utilix
        import pymongo

        contexts = utilix.xent_collection("contexts")
        ctx = contexts.find_one(
            {f"hashes.{self.data_type}": self.lineage_hash},
            projection={"context": "$name", "env": "$tag", "_id": 0},
            sort=[("date_added", pymongo.DESCENDING)],
        )

        return dict(ctx)



import rframe
from requests import PreparedRequest
from requests.auth import AuthBase


class ApiAuth(AuthBase):
    authenticator = None

    def __init__(self, authenticator=None) -> None:
        self.authenticator = authenticator

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        if self.authenticator is not None:
            r.headers["Authorization"] = f"Bearer {self.authenticator.token}"
        return r


def api_client(url, token=None, authenticator=None, headers=None, client=None):
    headers = headers if headers is not None else {}

    if token is not None:
        auth = f"Bearer {token}"
    elif authenticator is not None:
        auth = ApiAuth(authenticator)
    else:
        auth = None
    client = rframe.RestClient(
        url,
        headers=headers,
        client=client,
        auth=auth,
    )
    return client



from collections import defaultdict


from rframe import DataAccessor
from xedocs import all_schemas
from xedocs.storage import ApiStorage, DictStorage, UtilixStorage
from xedocs.xedocs import find_schema

from .utils import DatasetCollection
from .schemas import XeDoc

from ._settings import settings


class XedocsContext:
    _schemas: dict
    storage: list
    _accessors: dict

    def __init__(self, schemas: dict = None, storage=None, by_category=True):
        if by_category:
            self._accessors = defaultdict(dict)
        else:
            self._accessors = {}

        if storage is None:
            storage = []
        self.storage = storage

        if schemas is None:
            schemas = all_schemas()

        for name, schema in schemas.items():
            if by_category:
                category = schema._CATEGORY
            else:
                category = None
            self.register_schema(schema, name=name, category=category)

    def register_schema(self, schema, name=None, category=None):
        if isinstance(schema, str):
            schema = find_schema(schema)

        if not issubclass(schema, XeDoc):
            raise TypeError(
                "Wrong type for schema spec, "
                f"expected a XeDoc or string got {type(schema)}"
            )

        if name is None:
            name = schema._ALIAS

        datasource = self.get_accessor(schema, name=name)

        if datasource is None:
            raise ValueError(f"No datasource found for {name}")

        if category is None:
            self._accessors[name] = datasource
        else:
            self._accessors[category][name] = datasource

    def get_accessor(self, schema, name=None):
        if name is None:
            name = schema._ALIAS
        for store in self.storage:
            datasource = store.get_datasource(name)
            if datasource is not None:
                return DataAccessor(schema, datasource)

    def __getitem__(self, key):
        dset = self._accessors.get(key, None)
        if dset is None:
            raise KeyError(f"No dataset named {key} found")
        if isinstance(dset, dict):
            return DatasetCollection(dset)
        return self._accessors.get(key)

    def __getattr__(self, attr):
        if attr in self._accessors.keys():
            return self[attr]
        raise AttributeError(attr)

    def __dir__(self):
        return super().__dir__() + list(self._accessors.keys())


def straxen_db(schemas=None, datasource_overrides=None, by_category=False):
    if schemas is None:
        schemas = all_schemas()

    storage = [
        DictStorage(datasource_overrides),
        UtilixStorage(database=settings.STRAXEN_DB),
        ApiStorage(mode='straxen'),
    ]

    return XedocsContext(storage=storage, schemas=schemas, by_category=by_category)


def analyst_db(schemas=None, datasource_overrides=None, by_category=False):
    if schemas is None:
        schemas = all_schemas()

    storage = [
        DictStorage(datasource_overrides),
        UtilixStorage(database="xedocs"),
        ApiStorage(mode='analyst'),
    ]

    return XedocsContext(storage=storage, schemas=schemas, by_category=by_category)



"""
# Electron lifetime
Thanks to the continuous purification, we expect an improved electron lifetime over time. However, a significant number of variations are expected in the XENONnT detector (due to operations etc), we thus need a model to evaluate the evolution of e-life for time dependent S2 corrections. Thanks to purity monitor, we will have sufficient amount of data for such a study in XENONnT.

Currently we are using data from the purity monitor via SCADA interface to predict elife values in the next 6 hours, then this information is pass to corrections DB and updated there. (See e-log for updates)

See [description in the Team C overview page](https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:analysis:reconstruction_team#electron_lifetimeevolution_and_correction)

"""

from .base_corrections import TimeSampledCorrection


class ElectronLifetime(TimeSampledCorrection):
    _ALIAS = "electron_lifetimes"
    value: float



import rframe
import datetime

from .base_schemas import VersionedXeDoc
from .constants import PARTITION


class DetectorNumber(VersionedXeDoc):
    """Detector parameters
    A collection of non-time dependent detector
    values.
    """

    _ALIAS = "detector_numbers"

    field: str = rframe.Index(max_length=80)
    partition: PARTITION = rframe.Index(default="all_tpc")

    value: float
    uncertainty: float
    definition: str
    reference: str = ""
    date: datetime.datetime
    comments: str = ""



"""
# Correction: Electron diffusion
# What is it correction: S2 width cut (cutax)
# Description: This correction is to take into account how wide out S2s can be as higher diffusion implies longer S2s. While we do not directly correct our waveforms with this information it is used in cutax for the cut_s2_width cut.
# Latest wiki reference: 
"""

import rframe

from .base_corrections import TimeSampledCorrection


class ElectronDiffusionCte(TimeSampledCorrection):

    _ALIAS = "electron_diffusion_cte"
    value: float



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tests',
 'xedocs',
 'xedocs.schemas',
 'xedocs.schemas.analysis',
 'xedocs.schemas.calibrations',
 'xedocs.schemas.corrections',
 'xedocs.schemas.operations_reports',
 'xedocs.schemas.pmt_data']

package_data = \
{'': ['*']}

install_requires = \
['click', 'rframe>=0.2.0,<0.3.0', 'rich']

extras_require = \
{'db': ['pymongo']}

entry_points = \
{'console_scripts': ['xedocs = xedocs.cli:main'],
 'straxen': ['_ = xedocs._straxen_plugin']}

setup_kwargs = {
    'name': 'xedocs',
    'version': '0.2.5',
    'description': 'Top-level package for xedocs.',
    'long_description': "=======================================\nXeDocs - XENON Metadata management tool\n=======================================\nxedocs is meant to replace cmt and bodega as well as helping tracking all shared documents especially if\nthey need to be versioned.\n\nWhat does Xedocs give you\n=========================\n\nData reading\n------------\n\n- Read data from multiple formats (e.g. mongodb, pandas) and locations with a simple unified interface.\n- Custom logic implemented on the document class, e.g. creating a tensorflow model from the data etc.\n- Multiple APIs for reading data, fun functional, ODM style, pandas and xarray.\n- Read data as objects, dataframes, dicts, json.\n    \nWriting data\n------------\n\n- Write data to multiple storage backends with the same interface\n- Custom per-collection rules for data insertion, deletion and updating.\n- Schema validation and type coercion so storage has uniform and consistent data.\n    \nOther\n-----\n\n- Custom panel widgets for graphical representation of data, web client\n- Auto-generated API server and client + openapi documentation\n- CLI for viewing and downloading data\n\n\nBasic Usage\n-----------\n\nExplore the available schemas\n\n.. code-block:: python\n\n    import xedocs\n\n    >>> xedocs.list_schemas()\n    >>> ['detector_numbers',\n        'fax_configs',\n        'plugin_lineages',\n        'context_lineages',\n        'pmt_gains',\n        'global_versions',\n        'electron_drift_velocities',\n        ...]\n\n    >>> xedocs.help('pmt_gains')\n\n    >>>\n            Schema name: pmt_gains\n            Index fields: ['version', 'time', 'detector', 'pmt']\n            Column fields: ['created_date', 'comments', 'value']\n    \n\nRead/write data from the shared analyst database, this database is writable from the default analysis username/password\n\n.. code-block:: python\n\n    import xedocs\n\n    db = xedocs.analyst_db()\n\n    docs = db.pmt_gains.find_docs(version='v1', pmt=[1,2,3,5], time='2021-01-01T00:00:00', detector='tpc')\n    gains = [doc.value for doc in docs]\n\n    doc = db.pmt_gains.find_one(version='v1', pmt=1, time='2021-01-01T00:00:00', detector='tpc')\n    pmt1_gain = doc.value\n\nRead from the straxen processing database, this database is read-only for the default analysis username/password\n\n\n.. code-block:: python\n\n    import xedocs\n\n    db = xedocs.straxen_db()\n\n    ...\n    \nYou can also query documents directly from the schema class, \nSchemas will query the mongodb analyst database by default, if no explicit datasource is given.\n\n.. code-block:: python\n\n    from xedocs.schemas import DetectorNumber\n\n    drift_velocity = DetectorNumber.straxen_db.find_one(field='drift_velocity', version='v1')\n    \n    # Returns a Bodega object with attributes value, description etc.\n    drift_velocity.value\n\n    all_v1_documents = DetectorNumber.straxen_db.find(version='v1')\n\n\n\nRead data from alternative data sources specified by path, \ne.g csv files which will be loaded by pandas.\n\n.. code-block:: python\n\n    from xedocs.schemas import DetectorNumber\n    \n    g1_doc = DetectorNumber.find_one(datasource='/path/to/file.csv', version='v1', field='g1')\n    g1_value = g1_doc.value\n    g1_error = g1_doc.uncertainty\n\nThe path can also be a github URL or any other URL supported by fsspec. \n\n.. code-block:: python\n\n    from xedocs.schemas import DetectorNumber\n    \n    g1_doc = DetectorNumber.find_one(\n                             datasource='github://org:repo@/path/to/file.csv', \n                             version='v1', \n                             field='g1')\n\n\nSupported data sources\n\n    - MongoDB collections\n    - TinyDB tables\n    - JSON files\n    - REST API clients\n\nPlease open an issue on rframe_ if you want support for an additional data format.\n\n\nDocumentation\n-------------\nFull documentation hosted by Readthedocs_\n\nCredits\n-------\n\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n.. _Readthedocs: https://xedocs.readthedocs.io/en/latest/\n.. _rframe: https://github.com/jmosbacher/rframe",
    'author': 'Yossi Mosbacher',
    'author_email': 'joe.mosbacher@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/XENONnT/xedocs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""Top-level package for xedocs."""

__author__ = """Yossi Mosbacher"""
__email__ = "joe.mosbacher@gmail.com"
__version__ = "0.2.5"
import logging

logger = logging.getLogger(__name__)

from . import api
from ._settings import settings
from .datasources import *
from . import schemas
from .utils import *
from .xedocs import *

from .contexts import *

try:
    from . import _straxen_plugin

except ImportError:
    logger.warning("Could not register straxen protocol, \
                     most likely straxen not installed.")
except ValueError:
    pass

try:
    from . import widgets

    gui = widgets.XedocsEditor()
except ImportError:
    logger.warning("Could not import editors, GUI not available.")




from rframe import BaseSchema

from ._settings import settings, uconfig, xent_collection


def api_client(schema, mode='analyst'):
    import xedocs

    url = settings.api_url_for_schema(schema, mode=mode)

    return xedocs.api.api_client(url, token=settings.API_TOKEN, authenticator=settings)


def utilix_datasource(schema: BaseSchema, mode='analyst'):
    if uconfig is None:
        return None

    database = getattr(schema, f'__{mode.capitalize()}_DB__', None)
    if database is None:
        database = getattr(settings, f"{mode.capitalize()}_DB", "xedocs")
    collection = schema.default_collection_name()
    return xent_collection(collection=collection, database=database)


def default_datasource(schema, mode='analyst'):
    datasource = utilix_datasource(schema, mode=mode)
    if datasource is None:
        datasource = api_client(schema, mode=mode)
    return datasource        


def get_datasource_for(schema, mode='analyst'):
    if schema._ALIAS in settings.datasources:
        return settings.datasources[schema._ALIAS]

    return default_datasource(schema, mode=mode)


def register_default_storage(schema: BaseSchema):
    if uconfig is not None:
        if not hasattr(schema, "analyst_db"):
            try:
                schema.register_datasource(utilix_datasource(schema), name='analyst_db')
            except:
                pass
        if not hasattr(schema, "straxen_db"):
            try:
                schema.register_datasource(utilix_datasource(schema, mode='straxen'), name='straxen_db')
            except:
                pass

    if not hasattr(schema, "analyst_db_api"):
        try:
            schema.register_datasource(api_client(schema), name='analyst_db_api')
        except:
            pass

    if not hasattr(schema, "straxen_db_api"):
        try:
            schema.register_datasource(api_client(schema, mode='straxen'), name='straxen_db_api')
        except:
            pass



import io
import math
import json
import xedocs
import rframe
import datetime
import pydantic
import logging

import numpy as np
import pandas as pd
import panel as pn

import param

from typing import Any, ClassVar, List, Type

try:
    from typing import _LiteralGenericAlias
except ImportError:
    _LiteralGenericAlias = None

from panel.widgets import (
    LiteralInput,
    CompositeWidget,
    DatetimeRangePicker,
    EditableRangeSlider,
)

from panel.layout import ListPanel, Column
from panel.io.server import unlocked
from tornado.ioloop import IOLoop

from pydantic_panel import infer_widget, PydanticModelEditor, ItemListEditor

from pydantic_panel.dispatchers import clean_kwargs

from typing import ClassVar, Type


from concurrent.futures import ThreadPoolExecutor

from .schemas import XeDoc
from .utils import docs_to_wiki, docs_to_dataframe
from .dispatchers import json_serializable

executor = ThreadPoolExecutor(max_workers=5)  # pylint: disable=consider-using-with
logger = logging.getLogger(__name__)


def make_csv(schema, docs, fields=None):
    df = docs_to_dataframe(schema, docs, columns=fields)
    return io.StringIO(df.to_csv())


def make_json(schema, docs, fields=None):
    if fields is not None:
        fields = set(fields)
    docs = [json.loads(doc.json(include=fields)) for doc in docs]

    return io.StringIO(json.dumps(docs, indent=4))


def make_wiki(schema, docs, fields=None):
    return io.StringIO(docs_to_wiki(schema, docs, columns=fields))


download_options = {
    "csv": make_csv,
    "json": make_json,
    "wiki": make_wiki,
}


class NullableInput(LiteralInput):
    def _process_property_change(self, msg):
        # msg = Widget._process_property_change(self, msg)
        if "value" in msg and msg["value"] == "":
            msg["value"] = "null" if self.serializer == "json" else "None"

        msg = super()._process_property_change(msg)
        return msg


class TimeIntervalEditor(DatetimeRangePicker):
    value = param.ClassSelector(rframe.TimeInterval, default=None)

    def _serialize_value(self, value):
        value = super()._serialize_value(value)
        if any([v is None for v in value]):
            return None
        return rframe.TimeInterval(left=value[0], right=value[1])

    def _deserialize_value(self, value):
        if isinstance(value, rframe.TimeInterval):
            value = value.left, value.right

        value = super()._deserialize_value(value)
        return value

    @param.depends("start", "end", watch=True)
    def _update_value_bounds(self):
        pass


@infer_widget.dispatch(precedence=2)
def infer_widget(value: rframe.TimeInterval, field: Any, **kwargs):
    kwargs = clean_kwargs(TimeIntervalEditor, kwargs)
    return TimeIntervalEditor(value=value, **kwargs)


class IntegerIntervalEditor(EditableRangeSlider):

    step = param.Integer(default=1, constant=True)

    format = param.String(default="0", constant=True)

    value = param.ClassSelector(rframe.IntegerInterval, default=None)

    value_throttled = param.ClassSelector(rframe.IntegerInterval, default=None)

    @param.depends("value", watch=True)
    def _update_value(self):
        if self.value is None:
            return

        self._slider.value = (self.value.left, self.value.right)
        self._start_edit.value = self.value.left
        self._end_edit.value = self.value.right

    def _sync_value(self, event):
        with param.edit_constant(self):
            new_value = rframe.IntegerInterval(left=event.new[0], right=event.new[1])
            self.param.update(**{event.name: new_value})

    def _sync_start_value(self, event):
        if event.name == "value":
            end = self.value.right if self.value else self.end
        else:
            end = self.value_throttled.right if self.value_throttled else self.end

        new_value = rframe.IntegerInterval(left=event.new, right=end)

        with param.edit_constant(self):
            self.param.update(**{event.name: new_value})

    def _sync_end_value(self, event):
        if event.name == "value":
            start = self.value.left if self.value else self.start
        else:
            start = self.value_throttled.left if self.value_throttled else self.start

        new_value = rframe.IntegerInterval(left=start, right=event.new)
        with param.edit_constant(self):
            self.param.update(**{event.name: new_value})


@infer_widget.dispatch(precedence=2)
def infer_widget(value: rframe.IntegerInterval, field: Any, **kwargs):
    start = None
    end = None

    if start is None:
        start = kwargs.get("start", value._min if value is not None else 0)

    if end is None:
        end = kwargs.get("end", value._max if value is not None else 1)

    step = value._resolution if value else 1

    kwargs = clean_kwargs(IntegerIntervalEditor, kwargs)
    return IntegerIntervalEditor(value=value, step=step, start=start, end=end, **kwargs)


class XeDocEditor(PydanticModelEditor):

    _trigger_recreate = ["class_", "allow_save", "allow_delete"]

    datasource = param.Parameter(default=None)
    class_ = param.ClassSelector(class_=XeDoc, is_instance=False)
    value = param.ClassSelector(class_=XeDoc)

    pre_save = param.HookList([])
    post_save = param.HookList([])
    pre_delete = param.HookList([])
    post_delete = param.HookList([])

    delete_requested = param.Boolean(False)
    deleted = param.Boolean(False)
    deletion_error = param.String("")

    allow_save = param.Boolean(True)

    allow_delete = param.Boolean(True)

    add_debugger = param.Boolean(False)

    def __init__(self, **params):
        if params.get("fields", None) is None:
            schema = params.get("class_", None) or params.get("value", None)
            if schema is not None:
                index_fields = list(schema.get_index_fields())
                column_fields = list(schema.get_column_fields())
                params["fields"] = index_fields + column_fields

        super().__init__(**params)

    def _update_save_button(self, target, event):
        target.disabled = event.new is None

    @property
    def can_delete(self):
        if self.deleted or self.value is None:
            return False
        try:
            self.value.pre_delete(self.datasource)
            return True
        except:
            return False

    def _save_clicked(self, event):
        if self.value is None:
            return

        event.obj.name = "Saving..."
        self.loading = True
        try:
            doc = self.value
            for cb in self.pre_save:
                cb(doc)

            doc.save(self.datasource)

            for cb in self.post_save:
                cb(doc)
        finally:
            self.loading = False
            event.obj.name = "Save"

    @pn.depends("deleted", "delete_requested")
    def delete_button(self):
        if self.deleted:
            return pn.widgets.Button(name="Deleted.", disabled=True)

        if self.delete_requested:
            cancel = pn.widgets.Button(
                name="❌ Cancel",
            )

            def cancel_cb(event):
                self.delete_requested = False

            cancel.on_click(cancel_cb)

            confirm = pn.widgets.Button(
                name="⚠️ Confirm? This cannot be undone. ⚠️",
                button_type="danger",
                width_policy="max",
            )

            def confirm_cb(event):
                confirm.disabled = True
                try:
                    self.delete()
                finally:
                    confirm.disabled = False

            confirm.on_click(confirm_cb)

            return pn.Column(cancel, confirm)

        delete = pn.widgets.Button(
            name="Delete 🗑️", button_type="danger", width_policy="max"
        )

        def delete_cb(event):
            self.delete_requested = True

        delete.on_click(delete_cb)

        if self.deletion_error:
            error_message = pn.pane.Alert(self.deletion_error, alert_type="danger")
            acknowledge = pn.widgets.Button(
                name="✔️ OK",
            )

            def ack_cb(event):
                self.deletion_error = ""

            acknowledge.on_click(ack_cb)

            return pn.Column(error_message, acknowledge)

        return delete

    def delete(self):
        self.loading = True
        try:
            doc = self.value

            for cb in self.pre_delete:
                cb(doc)

            doc.delete(self.datasource)
            self.deleted = True

            for cb in self.post_delete:
                cb(doc)
        except Exception as e:
            self.deletion_error = str(e)
            logger.error(str(e))
        finally:
            self.loading = False

    def _add_buttons(self):

        if self.allow_save:
            save_button = pn.widgets.Button(
                name="Save 💾", button_type="success", disabled=self.value is None
            )
            self.link(save_button, callbacks={"value": self._update_save_button})
            save_button.on_click(self._save_clicked)
            self._composite.append(save_button)

        if self.allow_delete and self.can_delete:
            self._composite.append(self.delete_button)

        if self.add_debugger:
            debugger = pn.widgets.Debugger(name="Debugger", width_policy="max")
            self._composite.append(debugger)

    def _recreate_widgets(self, *events):
        super()._recreate_widgets(*events)
        self._add_buttons()


@infer_widget.dispatch(precedence=2)
def infer_widget(value: XeDoc, field: Any, **kwargs):
    if field is not None:
        kwargs["name"] = kwargs.pop("name", field.name)
    kwargs = clean_kwargs(XeDocEditor, kwargs)
    return XeDocEditor(value=value, **kwargs)


class QueryEditor(CompositeWidget):

    _composite_type = pn.Column

    _trigger_recreate: ClassVar[List] = ["class_", "extra_widgets", "fields"]

    class_ = param.ClassSelector(XeDoc, is_instance=False)

    value = param.Dict({})

    _widgets = param.Dict({})

    fields = param.List()

    extra_widgets = param.List([])

    by_alias = param.Boolean(False)

    def __init__(self, **params):
        super().__init__(**params)
        self._update_layout()
        self.param.watch(self._update_layout, self._trigger_recreate)

    def _widget_for(self, field_name):
        field = self.class_.__fields__[field_name]

        alias = field.alias if self.by_alias else field_name

        alias = alias.replace("_", " ").capitalize()

        if type(field.outer_type_) == _LiteralGenericAlias:
            options = list(field.outer_type_.__args__)
            widget = pn.widgets.MultiChoice(name=alias + "s", value=[], options=options)
        else:
            try:
                widget = ItemListEditor(
                    value=[],
                    class_=field.outer_type_,
                    item_field=field,
                    name=alias + "s",
                )
            except:
                widget = NullableInput(
                    value=None,
                    name=alias,
                )

        widget.param.watch(self._validate_field, "value")
        return widget

    def _make_widgets(self):
        fields = self.fields or list(self.class_.get_index_fields())
        self._widgets = {
            field_name: self._widget_for(field_name) for field_name in fields
        }

    def _update_layout(self, *events):
        if self.class_ is None:
            return

        self._make_widgets()

        self._composite[:] = list(self._widgets.values())

        for w in self.extra_widgets:
            self._composite.append(w)

    def _validate_field(self, event):

        for name, widget in self._widgets.items():
            if event.obj == widget:
                break
        else:
            return

        index = self.class_.index_for(name)

        label = index.validate_label(event.new)

        if label:
            self.value[name] = label
        else:
            self.value[name] = None

        self.param.trigger("value")


class XeDocListEditor(CompositeWidget):
    """Composite widget for editing a list of XeDocs"""

    _composite_type: ClassVar[Type[ListPanel]] = Column

    row_size = param.Integer(30)
    MAX_DISPLAY_ROWS = 20
    MAX_SEND_ROWS = 500
    column_size = param.Integer(100)

    page_size = param.Integer(15)

    class_ = param.ClassSelector(XeDoc, is_instance=False)

    pagination = param.ObjectSelector(
        default=None, allow_None=True, objects=["local", "remote"]
    )

    value = param.List(default=[], class_=XeDoc)

    table_widget = param.Parameter(default=None)

    def __init__(self, **params):
        super().__init__(**params)
        if self.class_ is None and self.value:
            self.class_ = self.value[0].__class__
        self._composite[:] = [pn.panel(self.table_view)]

    @param.depends("class_", "pagination", "page_size", watch=True, on_init=True)
    def _make_table_widget(self):

        if len(self.value) > self.MAX_SEND_ROWS:
            self.pagination = "remote"

        elif len(self.value) > self.MAX_DISPLAY_ROWS:
            self.pagination = "local"

        docs = [json_serializable(doc.index_labels) for doc in self.value]

        if self.class_ is not None:
            df = pd.DataFrame(docs, columns=list(self.class_.get_index_fields()))
            nwidgets = len(self.class_.get_column_fields()) + 2
            ncols = len(self.class_.get_index_fields()) + 2

        else:
            df = pd.DataFrame()
            nwidgets = 2
            ncols = 3

        nrows = max(self.MAX_DISPLAY_ROWS, nwidgets) + 1

        self.table_widget = pn.widgets.Tabulator(
            df,
            name="Data (Click to edit)",
            disabled=True,
            row_content=self._value_editor,
            sizing_mode="stretch_both",
            min_height=nrows * self.row_size,
            min_width=ncols * self.column_size,
            width_policy="max",
            height_policy="max",
            embed_content=False,
            page_size=self.page_size,
            pagination=self.pagination,
            show_index=False,
        )

    @param.depends(
        "value",
        watch=True,
    )
    def _update_table(self):

        if len(self.value) > self.MAX_SEND_ROWS:
            self.pagination = "remote"

        elif len(self.value) > self.MAX_DISPLAY_ROWS:
            self.pagination = "local"

        if self.table_widget is None:
            return

        if not len(self.table_widget.value) or not self.value:
            return self._make_table_widget()

        assert all(
            [isinstance(doc, self.class_) for doc in self.value]
        ), f"Value must be a list of {self.class_} instances"

        docs = [json_serializable(doc.index_labels) for doc in self.value]

        df = pd.DataFrame(docs, columns=list(self.class_.get_index_fields()))
        self.table_widget.value = df

    @pn.depends("table_widget")
    def table_view(self):
        if self.table_widget is None:
            return pn.Column()
        return pn.Column(self.table_widget)

    def _value_editor(self, row):
        if len(self.value) <= row.name:
            return pn.Column()

        doc = self.value[row.name]

        columns = list(doc.get_column_fields())

        editor = XeDocEditor(
            value=doc,
            class_=self.class_,
            fields=columns,
            allow_delete=True,
        )

        def cb(event):
            if event.new:
                self.value.pop(row.name, None)
                self.param.trigger("value")

        editor.param.watch(cb, "deleted")

        return pn.Column(editor)


@infer_widget.dispatch(precedence=2)
def infer_widget(value: List[XeDoc], field, **kwargs):
    if field is not None:
        kwargs["class_"] = kwargs.pop("class_", field.type_)

    kwargs = clean_kwargs(XeDocListEditor, kwargs)

    return XeDocListEditor(value=value, **kwargs)


class ModelTableEditor(pn.viewable.Viewer):

    class_ = param.ClassSelector(pydantic.BaseModel, is_instance=False)

    docs = param.List(class_=pydantic.BaseModel)

    query_editor = param.ClassSelector(QueryEditor)
    model_editor = param.ClassSelector(XeDocEditor)

    query = param.Dict({})

    page = param.Integer(-1, bounds=(-1, 1))
    last_page = param.Integer(0)
    page_size = param.Integer(15)

    refresh_table = param.Event()
    table = param.Parameter()

    def __init__(self, **params):
        super().__init__(**params)

        self.inc_page = pn.widgets.Button(
            name="\u25b6",
            disabled=(self.page >= self.last_page or self.page == -1),
            height=50,
            max_width=50,
            align="center",
        )
        self.inc_page.on_click(self.increment_page)

        self.dec_page = pn.widgets.Button(
            name="\u25c0",
            disabled=self.page <= 0,
            align="center",
            height=50,
            width=50,
        )
        self.dec_page.on_click(self.decrement_page)

        self.find_button = pn.widgets.Button(
            name="🔍 Query", button_type="primary", align="center"
        )

        self.find_button.on_click(self.filter_callback)

        self.param.watch(
            self._update_docs, ["class_", "page", "query", "refresh_table"]
        )

    @param.depends("class_", watch=True, on_init=True)
    def _make_table_widget(self):
        docs = [json_serializable(doc.index_labels) for doc in self.docs]
        df = pd.DataFrame(docs, columns=list(self.class_.get_index_fields()))

        self.table = pn.widgets.Tabulator(
            df,
            name="Data (Click to edit)",
            disabled=True,
            row_content=self.value_editor,
            sizing_mode="stretch_both",
            embed_content=False,
            show_index=False,
            #  width=1000,
            min_height=500,
        )

        self.query_editor = QueryEditor(class_=self.class_)

        self.model_editor = XeDocEditor(class_=self.class_, allow_delete=False)
        self.model_editor.post_save.append(self._update_docs)

    @param.depends("docs", watch=True)
    def _docs_changed(self):
        if self.table is None:
            self._make_table_widget()
            return

        if not len(self.table.value):
            return self._make_table_widget()

        docs = [json_serializable(doc.index_labels) for doc in self.docs]
        df = pd.DataFrame(docs, columns=list(self.class_.get_index_fields()))
        self.table.value = df
        self.table.loading = False

    @param.depends("page", watch=True)
    def _page_changed(self):
        self.inc_page.disabled = self.page >= self.last_page or self.page == -1
        self.dec_page.disabled = self.page <= 0

    @param.depends("last_page", watch=True)
    def _last_page_changed(self):
        self.inc_page.disabled = self.page >= self.last_page or self.page == -1

    def filter_callback(self, event):
        query = {k: v for k, v in self.query_editor.value.items() if v}
        query = {k: v[0] if len(v) == 1 else v for k, v in query.items()}

        if self.query != query:
            self.query = query
            self.page = 0
        if self.page == -1:
            self.page = 0
        # end = math.ceil(self.class_.compile_query(**self.query).count()/self.page_size)
        # self.param.page.bounds = (0, end)
        # self.last_page = end
        self._update_page_range()
        self._update_docs()

    def trigger_refresh_cb(self, event):
        self.refresh_table = True

    def increment_page(self, event):
        self.page += 1

    def decrement_page(self, event):
        self.page -= 1

    def _set_page_range(self, future):
        count = future.result()
        last_page = math.ceil(count / self.page_size)
        with unlocked():
            self.param.page.bounds = (0, last_page)
            self.last_page = last_page

    def _set_docs(self, future):
        docs = future.result()

        if docs == self.docs:
            self.table.loading = False

        with unlocked():
            self.docs = docs
            docs = [json_serializable(doc.index_labels) for doc in self.docs]

    def _get_page(self):
        skip = self.page * self.page_size
        docs = []
        try:
            docs = self.class_.find(**self.query, skip=skip, limit=self.page_size)
        except Exception as e:
            logger.error(str(e))

        return docs

    def _get_page_count(self):
        count = 1
        try:
            count = self.class_.compile_query(**self.query).count()
        except Exception as e:
            logger.error(str(e))
        return count

    def _update_docs(self, *events):
        self.table.loading = True
        loop = IOLoop.current()
        future = executor.submit(self._get_page)
        loop.add_future(future, self._set_docs)

    def _update_page_range(self, *events):
        self.table.loading = True
        loop = IOLoop.current()
        future = executor.submit(self._get_page_count)
        loop.add_future(future, self._set_page_range)

    def value_editor(self, row):
        if len(self.docs) <= row.name:
            return pn.Column()

        doc = self.docs[row.name]

        columns = list(doc.get_column_fields())

        editor = XeDocEditor(
            value=doc,
            class_=self.class_,
            fields=columns,
            allow_delete=True,
        )

        editor.post_delete.append(self._update_docs)

        return pn.Column(editor)

    @pn.depends("table")
    def table_panel(self):
        if self.table is None:
            return pn.Column()
        return pn.Column(
            self.table,
            sizing_mode="stretch_width",
            #  width=1000,
            scroll=False,
        )

    @pn.depends("page", "last_page")
    def page_controls(self):
        end = self.last_page or 1

        page_view = pn.indicators.LinearGauge(
            name="Page",
            value=self.page,
            bounds=(0, end),
            format="{value}",
            horizontal=True,
            width=75,
            min_height=75,
            align="end",
            tick_size="12px",
            title_size="12px",
            value_size="12px",
        )
        return pn.Row(
            self.dec_page, page_view, self.inc_page, max_width=600, width_policy="max"
        )

    @pn.depends("query_editor")
    def query_panel(self):
        if self.query_editor is None:
            return pn.Column()
        return pn.Column(self.query_editor)

    def controls_panel(self):
        query_container = pn.Card(
            "Valid JSON or python literals",
            self.query_panel,
            header=pn.Row("🔍 Filter Documents", pn.layout.Spacer(), width=300),
            collapsed=True,
        )

        add_new = pn.Card(
            self.new_doc_panel,
            name="insert_new",
            header=pn.Row("➕ New Document", pn.layout.Spacer(), width=300),
            collapsed=True,
        )

        return pn.Column(
            query_container,
            self.find_button,
            pn.layout.Divider(),
            add_new,
        )

    @pn.depends("model_editor")
    def new_doc_panel(self):
        if self.model_editor is None:
            return pn.Column()
        return pn.Column(self.model_editor, width_policy="min")

    @pn.depends("class_")
    def download_panel(self):

        current_only = pn.widgets.RadioBoxGroup(
            options={"Current page": 1, "Entire query": 0}, inline=True
        )
        filename = pn.widgets.TextInput(name="Filename", value=self.class_._ALIAS)

        filetype = pn.widgets.Select(name="Format", options=list(download_options))

        fields = pn.widgets.MultiChoice(
            name="Fields",
            value=list(self.class_.__fields__),
            options=list(self.class_.__fields__),
        )

        def cb():
            if current_only.value:
                docs = self.docs
            else:
                docs = self.class_.find(**self.query)

            return download_options[filetype.value](
                self.class_, docs, fields=fields.value
            )

        download_button = pn.widgets.FileDownload(
            filename=f"{self.class_._ALIAS}.{filetype.value}", callback=cb
        )

        def update_filename(*events):
            download_button.filename = f"{filename.value}.{filetype.value}"

        filename.param.watch(update_filename, "value")
        filetype.param.watch(update_filename, "value")

        return pn.Column(current_only, fields, filename, filetype, download_button)

    def __panel__(self):
        right_panel = pn.Column(self.page_controls, self.table_panel)
        left_panel = self.controls_panel
        return pn.Row(
            left_panel, right_panel, scroll=False, sizing_mode="stretch_width"
        )


class XedocsEditor(pn.viewable.Viewer):
    _schemas_by_category = xedocs.schemas_by_category()

    selection_layout = param.ClassSelector(
        class_=ListPanel, is_instance=False, default=pn.Row
    )

    category = param.Selector(objects=list(_schemas_by_category))
    collection = param.Selector(objects=[None])

    editor = param.ClassSelector(ModelTableEditor)

    def __init__(self, **params):
        all_schemas = xedocs.schemas_by_category()
        schemas = params.pop("schemas", None)
        if schemas is None:
            schemas = xedocs.list_schemas()

        # if isinstance(schemas, dict):
        #     self._all_schemas = schemas
        #

        categories = params.pop("categories", None)

        if categories is None:
            categories = list(xedocs.schemas_by_category())

        schemas_by_category = {}
        for category in categories:
            cat_schemas = all_schemas.get(category, {})
            schema_dict = {k: v for k, v in cat_schemas.items() if k in schemas}
            schemas_by_category[category] = schema_dict

        self._schemas_by_category = schemas_by_category
        self.param.category.objects = list(schemas_by_category)
        super().__init__(**params)

    @param.depends("category", watch=True, on_init=True)
    def _category_changed(self):
        schemas = self._schemas_by_category.get(self.category, {})
        if not schemas:
            return

        objects = list(schemas.values())
        self.param.collection.names = schemas
        self.param.collection.objects = objects
        self.collection = objects[0]

    @param.depends("collection", watch=True, on_init=True)
    def _selection_changed(self):
        if self.collection is None:
            return
        self.editor = ModelTableEditor(class_=self.collection)

    @pn.depends("selection_layout")
    def selection_panel(self):
        return self.selection_layout(self.param.category, self.param.collection)

    @pn.depends("editor")
    def query_or_insert_panel(self):
        if self.editor is None:
            return pn.Column()
        return pn.panel(self.editor.controls_panel)

    @pn.depends("editor")
    def query_panel(self):
        if self.editor is None:
            return pn.Column()

        return pn.Column(
            "### Filters (python literals)", self.editor.query_panel, width_policy="max"
        )

    @pn.depends("editor")
    def new_doc_panel(self):
        if self.editor is None:
            return pn.Column()
        return pn.panel(self.editor.new_doc_panel)

    @pn.depends("editor")
    def data_panel(self):
        if self.editor is None:
            return pn.Column()
        return pn.Column(self.editor.page_controls, self.editor.table_panel)

    @pn.depends("editor")
    def find_button_panel(self):
        if self.editor is None:
            return pn.Row(pn.layout.Spacer(width_policy="max"))
        return pn.Row(
            pn.layout.Spacer(width_policy="max"),
            self.editor.find_button,
            pn.layout.Spacer(width_policy="max"),
        )

    @pn.depends("editor")
    def download_panel(self):
        if self.editor is None:
            return pn.Row(pn.layout.Spacer(width_policy="max"))
        return pn.panel(self.editor.download_panel)

    def __panel__(self):
        bottom_panel = pn.Row(self.query_or_insert_panel, self.data_panel)
        return pn.Column(self.selection_panel, pn.layout.Divider(), bottom_panel)
